/* Copyright 2020 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

import {PolymerElement, html} from '@polymer/polymer';
import {customElement, property} from '@polymer/decorators';

import '../../../components/polymer/irons_and_papers';
import * as tf_graph from '../tf_graph_common/graph';
import * as tf_graph_render from '../tf_graph_common/render';
import '../tf_graph_debugger_data_card/tf-graph-debugger-data-card';
import '../tf_graph_op_compat_card/tf-graph-op-compat-card';
import './tf-node-info';
import {LegacyElementMixin} from '../../../components/polymer/legacy_element_mixin';
import {ColorBy} from '../tf_graph_common/view_types';
import {Hierarchy} from '../tf_graph_common/hierarchy';

@customElement('tf-graph-info')
class TfGraphInfo extends LegacyElementMixin(PolymerElement) {
  static readonly template = html`
    <style>
      :host {
        background: var(--secondary-background-color);
        font-size: 12px;
        margin: 0;
        padding: 0;
        display: block;
        max-height: 650px;
        overflow-x: hidden;
        overflow-y: auto;
      }

      h2 {
        padding: 0;
        text-align: center;
        margin: 0;
      }
    </style>
    <template is="dom-if" if="{{selectedNode}}">
      <paper-material elevation="1" class="card">
        <tf-node-info
          graph-hierarchy="[[graphHierarchy]]"
          render-hierarchy="[[renderHierarchy]]"
          flat-graph="[[graph]]"
          graph-node-name="[[selectedNode]]"
          node-include="[[selectedNodeInclude]]"
          highlighted-node="{{highlightedNode}}"
          color-by="[[colorBy]]"
        >
        </tf-node-info>
      </paper-material>
    </template>
    <template is="dom-if" if="[[_equals(colorBy, 'op_compatibility')]]">
      <tf-graph-op-compat-card
        graph-hierarchy="[[graphHierarchy]]"
        render-hierarchy="[[renderHierarchy]]"
        color-by="[[colorBy]]"
        node-title="[[compatNodeTitle]]"
      >
      </tf-graph-op-compat-card>
    </template>
    <template
      is="dom-if"
      if="[[_healthPillsAvailable(debuggerDataEnabled, nodeNamesToHealthPills)]]"
    >
      <tf-graph-debugger-data-card
        render-hierarchy="[[renderHierarchy]]"
        debugger-numeric-alerts="[[debuggerNumericAlerts]]"
        node-names-to-health-pills="[[nodeNamesToHealthPills]]"
        selected-node="{{selectedNode}}"
        highlighted-node="{{highlightedNode}}"
        are-health-pills-loading="[[areHealthPillsLoading]]"
        all-steps-mode-enabled="{{allStepsModeEnabled}}"
        specific-health-pill-step="{{specificHealthPillStep}}"
        health-pill-step-index="{{healthPillStepIndex}}"
      >
      </tf-graph-debugger-data-card>
    </template>
  `;
  @property({type: String})
  override title: string;
  @property({type: Object})
  graphHierarchy: Hierarchy;
  @property({type: Object})
  graph: tf_graph.SlimGraph;
  @property({type: Object})
  renderHierarchy: tf_graph_render.RenderGraphInfo;
  @property({type: Object})
  nodeNamesToHealthPills: object;
  @property({
    type: Number,
    notify: true,
  })
  healthPillStepIndex: number;
  @property({type: String})
  colorBy: ColorBy;
  @property({type: String})
  compatNodeTitle: string;
  // Two-ways
  @property({
    type: String,
    notify: true,
  })
  selectedNode: string;
  /** @type {string?} */
  @property({
    type: String,
    notify: true,
  })
  highlightedNode: string;
  // The enum value of the include property of the selected node.
  @property({
    type: Number,
    notify: true,
  })
  selectedNodeInclude: number;
  // Whether debugger data is enabled for this instance of Tensorboard.
  @property({type: Boolean})
  debuggerDataEnabled: boolean;

  ready() {
    super.ready();

    this.addEventListener(
      'node-list-item-click',
      this._nodeListItemClicked.bind(this)
    );
    this.addEventListener(
      'node-list-item-mouseover',
      this._nodeListItemMouseover.bind(this)
    );
    this.addEventListener(
      'node-list-item-mouseout',
      this._nodeListItemMouseout.bind(this)
    );
  }
  _nodeListItemClicked(event) {
    this.selectedNode = event.detail.nodeName;
  }
  _nodeListItemMouseover(event) {
    this.highlightedNode = event.detail.nodeName;
  }
  _nodeListItemMouseout() {
    this.highlightedNode = null;
  }
  _healthPillsAvailable(debuggerDataEnabled, nodeNamesToHealthPills) {
    // So long as there is a mapping (even if empty) from node name to health pills, show the
    // legend and slider. We do that because, even if no health pills exist at the current step,
    // the user may desire to change steps, and the slider must show for the user to do that.
    return (
      debuggerDataEnabled &&
      nodeNamesToHealthPills &&
      Object.keys(nodeNamesToHealthPills).length > 0
    );
  }
  _equals(a, b) {
    return a === b;
  }
  // Whether health pills are currently being loaded, in which case we show a spinner (and the
  // current health pills shown might be out of date).
}
