/* Copyright 2020 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/
import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Input,
  OnDestroy,
  OnInit,
  Output,
} from '@angular/core';
import {Store} from '@ngrx/store';
import {DataLoadState} from '../../../types/data';
import {BehaviorSubject, combineLatest, Observable, Subject} from 'rxjs';
import {
  distinctUntilChanged,
  filter,
  map,
  shareReplay,
  takeUntil,
  tap,
} from 'rxjs/operators';

import {State} from '../../../app_state';
import {RunColorScale} from '../../../types/ui';
import * as actions from '../../actions';
import {
  ImageStepDatum,
  MetricsDataSource,
  PluginType,
} from '../../data_source/index';
import {
  getCardLoadState,
  getCardMetadata,
  getCardPinnedState,
  getCardStepIndex,
  getCardTimeSeries,
  getMetricsImageBrightnessInMilli,
  getMetricsImageContrastInMilli,
  getMetricsImageShowActualSize,
} from '../../store';
import {CardId, CardMetadata} from '../../types';
import {CardRenderer} from '../metrics_view_types';
import {getTagDisplayName} from '../utils';

type ImageCardMetadata = CardMetadata & {
  plugin: PluginType.IMAGES;
  sample: number;
  numSample: number;
  runId: string;
};

@Component({
  selector: 'image-card',
  template: `
    <image-card-component
      [loadState]="loadState$ | async"
      [title]="title$ | async"
      [tag]="tag$ | async"
      [runId]="runId$ | async"
      [sample]="sample$ | async"
      [numSample]="numSample$ | async"
      [imageUrl]="imageUrl$ | async"
      [stepIndex]="stepIndex$ | async"
      [stepValues]="stepValues$ | async"
      (stepIndexChange)="onStepIndexChanged($event)"
      [brightnessInMilli]="brightnessInMilli$ | async"
      [contrastInMilli]="contrastInMilli$ | async"
      [runColorScale]="runColorScale"
      [showActualSize]="showActualSize"
      [allowToggleActualSize]="(actualSizeGlobalSetting$ | async) === false"
      [isPinned]="isPinned$ | async"
      (onActualSizeToggle)="onActualSizeToggle()"
      (onPinClicked)="pinStateChanged.emit($event)"
    ></image-card-component>
  `,
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class ImageCardContainer implements CardRenderer, OnInit, OnDestroy {
  constructor(
    private readonly store: Store<State>,
    private readonly dataSource: MetricsDataSource
  ) {}

  @Input() cardId!: CardId;
  @Input() groupName!: string | null;
  @Input() runColorScale!: RunColorScale;
  @Output() fullWidthChanged = new EventEmitter<boolean>();
  @Output() pinStateChanged = new EventEmitter<boolean>();

  onStepIndexChanged($event: number) {
    this.store.dispatch(
      actions.cardStepSliderChanged({
        cardId: this.cardId,
        stepIndex: $event,
      })
    );
  }

  loadState$?: Observable<DataLoadState>;
  title$?: Observable<string>;
  tag$?: Observable<string>;
  runId$?: Observable<string>;
  sample$?: Observable<number>;
  numSample$?: Observable<number>;
  imageUrl$?: Observable<string | null>;
  stepIndex$?: Observable<number | null>;
  stepValues$?: Observable<number[]>;
  isPinned$?: Observable<boolean>;
  brightnessInMilli$ = this.store.select(getMetricsImageBrightnessInMilli);
  contrastInMilli$ = this.store.select(getMetricsImageContrastInMilli);
  actualSizeGlobalSetting$ = this.store.select(getMetricsImageShowActualSize);
  showActualSize = false;

  // The UI toggle is overridden by the global setting.
  private actualSizeUiToggled = false;
  private readonly actualSizeUiToggleSubject = new BehaviorSubject(
    this.actualSizeUiToggled
  );
  private readonly ngUnsubscribe = new Subject<void>();

  private isImageCardMetadata(
    cardMetadata: CardMetadata
  ): cardMetadata is ImageCardMetadata {
    const {plugin} = cardMetadata;
    return plugin === PluginType.IMAGES;
  }

  onActualSizeToggle() {
    this.actualSizeUiToggled = !this.actualSizeUiToggled;
    this.actualSizeUiToggleSubject.next(this.actualSizeUiToggled);
  }

  /**
   * Build observables once cardId is defined (after onInit).
   */
  ngOnInit() {
    combineLatest([
      this.actualSizeGlobalSetting$,
      this.actualSizeUiToggleSubject,
    ])
      .pipe(
        takeUntil(this.ngUnsubscribe),
        tap(([settingEnabled, uiToggleEnabled]) => {
          // Actual size mode requires a full width card.
          this.showActualSize = settingEnabled || uiToggleEnabled;
          this.fullWidthChanged.emit(this.showActualSize);
        })
      )
      .subscribe(() => {});

    const selectCardMetadata$ = this.store.select(getCardMetadata, this.cardId);
    const cardMetadata$ = selectCardMetadata$.pipe(
      takeUntil(this.ngUnsubscribe),
      filter((cardMetadata) => {
        return !!cardMetadata && this.isImageCardMetadata(cardMetadata);
      }),
      map((cardMetadata) => {
        return cardMetadata as ImageCardMetadata;
      }),
      shareReplay(1)
    );

    const metadataAndSeries$ = combineLatest([
      cardMetadata$,
      this.store.select(getCardTimeSeries, this.cardId),
    ]);
    const timeSeries$ = metadataAndSeries$.pipe(
      takeUntil(this.ngUnsubscribe),
      map(([cardMetadata, runToSeries]) => {
        const runId = cardMetadata.runId;
        if (!runToSeries || !runToSeries.hasOwnProperty(runId)) {
          return [];
        }
        return runToSeries[runId] as ImageStepDatum[];
      }),
      distinctUntilChanged((series1, series2) => {
        if (series1.length === series2.length && series1.length === 0) {
          return true;
        }
        return series1 === series2;
      }),
      shareReplay(1)
    );

    this.stepIndex$ = this.store.select(getCardStepIndex, this.cardId);
    this.loadState$ = this.store.select(getCardLoadState, this.cardId);

    const timeSeriesAndStepIndex$ = combineLatest([
      timeSeries$,
      this.stepIndex$,
    ]);
    const stepDatum$ = timeSeriesAndStepIndex$.pipe(
      map(([timeSeries, stepIndex]) => {
        if (stepIndex === null || !timeSeries[stepIndex]) {
          return null;
        }
        return timeSeries[stepIndex];
      })
    );

    this.tag$ = cardMetadata$.pipe(
      map((cardMetadata) => {
        return cardMetadata.tag;
      })
    );

    this.title$ = this.tag$.pipe(
      map((tag) => {
        return getTagDisplayName(tag, this.groupName);
      })
    );

    this.runId$ = cardMetadata$.pipe(
      map((cardMetadata) => {
        return cardMetadata.runId;
      })
    );

    this.sample$ = cardMetadata$.pipe(
      map((cardMetadata) => {
        return cardMetadata.sample;
      })
    );

    this.numSample$ = cardMetadata$.pipe(
      map((cardMetadata) => cardMetadata.numSample)
    );

    this.imageUrl$ = stepDatum$.pipe(
      map((stepDatum: ImageStepDatum | null) => {
        if (!stepDatum) {
          return null;
        }
        return this.dataSource.imageUrl(stepDatum.imageId);
      })
    );

    this.stepValues$ = timeSeries$.pipe(
      map((series: ImageStepDatum[]) => {
        return series.map((stepDatum) => {
          return stepDatum.step;
        });
      })
    );

    this.isPinned$ = this.store.select(getCardPinnedState, this.cardId);
  }

  ngOnDestroy() {
    this.ngUnsubscribe.next();
    this.ngUnsubscribe.complete();
  }
}
